(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
(*  Soubor: PRUNIK.PAS                                                     *)
(*  Obsah: program pro rozvinuti plaste valce protnuteho jinym valcem      *)
(*  Autor: Mircosoft (http://mircosoft.mzf.cz)                             *)
(*         Pouzity postup je prevzaty z Matematickeho fora:                *)
(*         http://forum.matweb.cz/viewtopic.php?id=7688                    *)
(*  Posledni uprava: 12.1.2021                                             *)
(*  Pro kompilaci: {$ifdef ukladani} CFG.TPU {$else} nic {$endif}          *)
(*  Pro spusteni: nic, ale hodi se mit po ruce obrazek SCHEMA.GIF          *)
(*  Upozorneni: tyto zdrojove kody pouzivate na vlastni nebezpeci          *)
(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
program pruniqalcu;
{$I+,N+} {zapnuti automaticke kontroly operaci se soubory (jsem liny to
         testovat rucne a pri pripadnem padu neni co zkazit)
         a zapnuti podpory matematickeho koprocesoru, aby se urychlily
         realne vypocty}

{$define ukladani} {definovano => pouzije se jednotka CFG a zadane hodnoty
                   se budou pamatovat pro priste
                   nedefinovano => program pujde zkompilovat bez jakychkoli
                   jednotek, ale hodnoty si pamatovat nebude a pri kazdem
                   spusteni je nastavi na vychozi}

{$ifdef ukladani}uses cfg;{$endif}

{$ifdef fpc}
{$packrecords 1}
{$endif}

const delkabufferu=8192; {po jak velkych [B] blocich se ma zapisovat do souboru (libovolne rozumne cislo ve wordovem rozsahu)}
      MaxBarva=3; {tohle+1=kolik barev bude v exportovanem obrazku (zvysovani nema smysl, pri zmenseni by byly nutne upravy)}

type HlavickaBMP = record {timhle musi zacinat kazdy soubor typu BMP}
                   znacka:array[1..2] of char; {'BM'}
                   VelikostSouboru:longint; {celeho vcetne vsech hlavicek, v B}
                   rezervovano:longint;
                   OfsetBitmapy:longint; {adresa prvniho bytu bitmapy, pocitano od 0 od zacatku souboru}
                   end;
     HlavickaDIBv3 = record {kompatibilni od Windows 3.0 nahoru, nejpouzivanejsi a hlavne se do ni da ulozit rozliseni}
                     VelikostHlavicky:longint; {velikost tehle hlavicky, vzdy 40 [B]}
                     sirka, {sirka obrazku [px]}
                     vyska:longint; {vyska obrazku [px]}
                     PocetBitovychRovin:word; {=1, vzdy}
                     BituNaPixel:word; {1, 4, 8, 16, 24 nebo 32, v nasem pripade 8}
                     komprese:longint; {0=zadna}
                     VelikostBitmapy:longint; {=vyska * zaokrouhlena sirka (nahoru na nejblizsi nasobek 4)}
                     rozliseniX,rozliseniY:longint; {[px/m]}
                     PocetBarevVPalete:longint; {aneb kolik ctyrbytovych polozek paleta obsahuje}
                     PocetPlatnychBarev:longint; {kolik barev je opravdu pouzitych, 0 = vsechny}
                     end;
     PaletaV3 = array[0..MaxBarva] of record
                                      B,G,R,P:byte; {hodnoty slozek 0..255, P je bud na pruhlednost nebo na nic, dame 0}
                                      end;

var {uzivatelske hodnoty (vsechny rozmery v mm a uhly v radianech):}
    r1,{polomer prvniho valce}
    l,{vyska prvniho valce}
    d,{vzdalenost prumetu pruseciku os obou valcu od cela prvniho valce}
    v,{vzdalenost os obou valcu}
    r2,{prumer druheho valce}
    {druhy valec uvazujeme nekonecne dlouhy}
    d1,d2,{prumery valcu}
    fi,{uhel mezi osami obou valcu}
    psi,{o kolik bude pootoceny spoj}
    tolerance:real;{jak tlustou chceme mit caru okraje pruniku:
                    0.5 - vetsinou 1 px, v nekterych mistech muze vyjit sirsi
                    <0.5 - v nejuzsich mistech uz se bude rozpadat
                    >0.5 - bude tlustsi
                    0 - nebude zadna}
    DPI:longint;{rozliseni, v jakem chceme exportovat [pixely na palec]}
    JmenoSouboru:string[79]; {cele jmeno souboru, do ktereho chceme exportovat
                              vysledek (vcetne koncovky BMP). Pozor: jestli
                              chcete zmenit maximalni delku, zmente ji i nize
                              v prikazu konf.definujpolozku('soubor'...}
    BarvaObrysu,{barva vnejsiho obrysu plaste valce}
    BarvaVnejsku,{barva povrchu prvniho valce}
    BarvaVnitrku,{barva plochy skryte ve druhem valci}
    BarvaOkraje:byte;{barva okraje pruniku}

    {pomocne promenne pro vypocty:}
    o, {pro pohyb po obvodu prvniho valce [px]}
    a:longint; {pro pohyb podel osy prvniho valce [px]}
    oMax,aMax:longint; {a a o nabyvaji hodnot od 0 do tohohle [px]}
    alfa,{uhlova souradnice zkoumaneho bodu (ma opacny smysl nez psi)}
    tgfi,{pro ulozeni hodnoty tg(fi), aby se nemusela pocitat vickrat}
    xA,yA,zA,{kartezske souradnice zkoumaneho bodu}
    DD,{posledni konstanta v obecne rovnici pomocne roviny}
    t,{parametr pro parametrickou rovnici osy druheho valce}
    xP,{yP=v,}zP,{souradnice paty kolmice od zkoumaneho bodu k ose druheho valce}
    vzdalenost:real;{delka teto kolmice}
    PravyUhel:boolean;{protoze pri fi=pi/2 musime pouzit specialni postup}

    {pomocne promenne pro praci se souborem a dalsi administrativu:}
    buffer:array[1..delkabufferu] of byte; {mezipamet pro blokovy zapis do souboru (aby se nemuselo psat byte po bytu)}
    PoziceVBufferu:word; {kolikaty byte v bufferu uz byl zaplnen (0=zatim zadny)}
    PoziceVRadku:longint; {kolikaty pixel aktualniho radku jsme prave dopsali (pri prazdnem radku 0)}
    soubor:file;
    hlavicka1:hlavickabmp;
    hlavicka2:hlavickadibv3;
    paleta:paletav3;
    b:byte; {na vykreslovani nabihajici listy}
    nacitadlo:string[79]; {pro nacitani veci z klavesnice}
{$ifdef ukladani}
    konf:konfigurace; {na ukladani uzivatelskych hodnot pro priste}
{$endif}

procedure ZapisBuffer; {vysype cely obsah bufferu do souboru}
Begin
blockwrite(soubor,buffer,pozicevbufferu);
pozicevbufferu:=0;
End;{zapisbuffer}

procedure ZapisPixel(barva:byte); {vlozi do bufferu jeden pixel}
Begin
if pozicevbufferu>=delkabufferu then zapisbuffer; {buffer je plny => do souboru s nim}
inc(pozicevbufferu);
buffer[pozicevbufferu]:=barva;
inc(pozicevradku);
End;{zapispixel}

function ZadejCislo(popis:string; vychozihodnota:real):real; {pro zadavani cisel z klavesnice}
var vysledek:real;
    kod:integer;
Begin
kod:=1;
 repeat
 if frac(vychozihodnota)=0 then write(popis,' <',vychozihodnota:0:0,'>: ')
                           else write(popis,' <',vychozihodnota:0:2,'>: ');
 readln(nacitadlo);
 if nacitadlo='' then zadejcislo:=vychozihodnota
                 else begin
                      val(nacitadlo,vysledek,kod);
                      if kod=0 then zadejcislo:=vysledek
                               else writeln('Chybne zadane cislo - opravte ho, prosim.');
                      end;
 until (nacitadlo='')or(kod=0);
End;{zadejcislo}

BEGIN
for b:=1 to 3 do writeln;
writeln('-      -     -    -   -  - Prunikator -  -   -    -     -      -');
writeln('   Program na rozvinuti plaste valce protnuteho jinym valcem');
writeln('****************************************************************');
writeln;
{vychozi hodnoty:}
d1:=40; d2:=20; l:=100; d:=40; v:=10; fi:=60; psi:=0;
jmenosouboru:='VYSTUP.BMP'; tolerance:=0; DPI:=96;
barvavnejsku:=1; barvaobrysu:=0; barvavnitrku:=2; barvaokraje:=3;
{$ifdef ukladani}
{nacteni minulych hodnot:}
cfgsoubor:='prunik.cfg';
konf.init;
konf.vybersekci('[vstupy]');
konf.definujpolozku('d1',addr(d1),_real);
konf.definujpolozku('d2',addr(d2),_real);
konf.definujpolozku('L',addr(l),_real);
konf.definujpolozku('D',addr(d),_real);
konf.definujpolozku('V',addr(v),_real);
konf.definujpolozku('fi',addr(fi),_real);
konf.definujpolozku('psi',addr(psi),_real);
konf.vybersekci('[vystupy]');
konf.definujpolozku('soubor',addr(jmenosouboru),_string+79);
konf.definujpolozku('tolerance okraje',addr(tolerance),_real);
konf.definujpolozku('DPI',addr(dpi),_longint);
konf.definujpolozku('barva plaste',addr(barvavnejsku),_byte);
konf.definujpolozku('barva obrysu plaste',addr(barvaobrysu),_byte);
konf.definujpolozku('barva diry',addr(barvavnitrku),_byte);
konf.definujpolozku('barva obrysu diry',addr(barvaokraje),_byte);
konf.nacti(cfgsoubor);
{$endif}
{prepocitani nactenych hodnot na pracovni format:}
r1:=d1/2; r2:=d2/2; fi:=fi*pi/180; psi:=psi*pi/180;
if barvavnejsku>3 then barvavnejsku:=1;
if barvaobrysu>3 then barvaobrysu:=0;
if barvavnitrku>3 then barvavnitrku:=2;
if barvaokraje>3 then barvaokraje:=3;

Repeat
writeln('Hlavni nabidka:');
writeln;
writeln(' samotny Enter = vypocet pruniku');
writeln(' N = navod k pouziti');
writeln(' cokoli jineho = konec');
writeln;
write('Vase volba: ');
readln(nacitadlo);
writeln;
if nacitadlo='' then {vypocet}
 begin
 {aktualizace vstupu:}
 writeln('Zadejte prosim vstupni hodnoty dle prilozeneho obrazku.');
 r1:=zadejcislo('Prumer d1 [mm]',2*r1)/2;
 r2:=zadejcislo('Prumer d2 [mm]',2*r2)/2;
 l:=zadejcislo('Delka L [mm]',l);
 d:=zadejcislo('Posunuti D [mm]',d);
 v:=zadejcislo('Vysunuti V [mm]',v);
 fi:=zadejcislo('Uhel fi [stupne]',fi*180/pi)*pi/180;
 while fi<0 do fi:=fi+pi; while fi>pi do fi:=fi-pi; {uprava na zakladni velikost}
 pravyuhel:=abs(fi-pi/2)<0.001;
 psi:=zadejcislo('Uhel spoje psi [stupne]',psi*180/pi)*pi/180;
 writeln;
 writeln('Zadejte prosim parametry vystupniho souboru:');
 writeln('(POZOR: pokud uz soubor se stejnym jmenem existuje, bude prepsan!)');
 write('Jmeno souboru (vcetne koncovky BMP) <',jmenosouboru,'>: ');
 readln(nacitadlo); if nacitadlo<>'' then jmenosouboru:=nacitadlo;
 DPI:=round(zadejcislo('DPI (rozliseni) [px/inch]',dpi));
 tolerance:=zadejcislo('Tolerance obrysu diry [px]',tolerance);
 writeln('Barvy: plast ',barvavnejsku,', obrys plaste ',barvaobrysu,
         ', dira ',barvavnitrku,', obrys diry ',barvaokraje);
 writeln('(moznosti: 0 = cerna, 1 = bila, 2 = svetleseda, 3 = tmavoseda)');
 write('Enter = nechat barvy, jak jsou; cokoli jineho + Enter = upravit ');
 readln(nacitadlo); if nacitadlo<>''
   then begin
        barvavnejsku:=round(zadejcislo('Barva plaste',barvavnejsku));
        barvaobrysu:=round(zadejcislo('Barva obrysu plaste',barvaobrysu));
        barvavnitrku:=round(zadejcislo('Barva diry',barvavnitrku));
        barvaokraje:=round(zadejcislo('Barva obrysu diry',barvaokraje));
        end;
 writeln;

 {--------------- vypocet: -----------------}

 writeln('Pocitam, cekejte prosim...');

 amax:=round(l*DPI/25.4)-1;
 omax:=round(2*pi*r1*DPI/25.4)-1;

 assign(soubor,jmenosouboru);
 rewrite(soubor,1);

 with hlavicka2 do begin
                   VelikostHlavicky:=40;
                   sirka:=omax+1;
                   vyska:=amax+1;
                   PocetBitovychRovin:=1;
                   BituNaPixel:=8; {co byte, to pixel}
                   komprese:=0; {bez komprese}
                   o:=sirka; while (o and 3)<>0 do inc(o); {(lokalni zneuziti o do funkce pomocne promenne)}
                   VelikostBitmapy:=o*vyska;
                   rozliseniX:=round(DPI*1000/25.4); {prevedeno z DPI (px/inch) na px/m}
                   rozliseniY:=round(DPI*1000/25.4);
                   PocetBarevVPalete:=maxbarva+1;
                   PocetPlatnychBarev:=0; {0 znamena vsechny}
                   end;
 with hlavicka1 do begin
                   znacka:='BM';
                   OfsetBitmapy:=sizeof(hlavickabmp)+sizeof(hlavickadibv3)+sizeof(paletav3);
                   VelikostSouboru:=ofsetbitmapy+hlavicka2.velikostbitmapy;
                   rezervovano:=0;
                   end;
 fillchar(paleta,sizeof(paletav3),0);
{with paleta[0] do begin b:=0;   g:=0;   r:=0;   end; {cerna}
 with paleta[1] do begin b:=255; g:=255; r:=255; end; {bila}
 with paleta[2] do begin b:=190; g:=190; r:=190; end; {svetleseda}
 with paleta[3] do begin b:=100; g:=100; r:=100; end; {tmavoseda}

 blockwrite(soubor,hlavicka1,sizeof(hlavickabmp));
 blockwrite(soubor,hlavicka2,sizeof(hlavickadibv3));
 blockwrite(soubor,paleta,sizeof(paletav3));

 pozicevbufferu:=0;

 if pravyuhel then zP:=d {kdyz je osa 2. valce kolma na osu z, bude jeji zova souradnice konstantni}
              else tgfi:=sin(fi)/cos(fi); {pri pravem uhlu by tg utekla do nekonecna}

 for a:=0 to amax do {cyklus pro kazdy pixel na plasti prvniho valce (a podel osy - svisle, o napric po obvodu - vodorovne)}
  begin
  if a and 63=0 then begin {vypsani ukazatele postupu (jenom jednou za 64 radku, aby moc nezdrzoval)}
                     write(#13'...........]'); {#13 = navrat na zacatek radku (CR)}
                     write(#13'[');
                     for b:=1 to a*longint(10) div amax do write('>');
                     end;
  pozicevradku:=0;
  zA:=a*25.4/DPI;
  for o:=0 to omax do
   begin
   {vypocet vzdalenosti zkoumaneho bodu od osy druheho valce
    (postup je vysvetlen v komentari za koncem programu):}
   alfa:=(o*25.4)/(r1*DPI)-psi;
   xA:=r1*cos(alfa);
   yA:=r1*sin(alfa);
   if pravyuhel then xP:=xA
                else begin
                     DD:=-tgfi*xA-zA;
                     t:=(sqr(tgfi)*d-DD)/(sqr(tgfi)+1);
                     xP:=tgfi*t-tgfi*d;
                     {yP:=v;}
                     zP:=t;
                     end;
   vzdalenost:=sqrt(sqr(xA-xP)+sqr(yA-v{yP})+sqr(zA-zP));
   {vybarveni pixelu podle vysledku vypoctu:}
   if (a=0)or(o=0)or(a=amax)or(o=omax) then zapispixel(barvaobrysu)
    else if abs(vzdalenost-r2)<=tolerance*25.4/DPI then zapispixel(barvaokraje)
     else if vzdalenost<r2 then zapispixel(barvavnitrku)
      else zapispixel(barvavnejsku);
   end;
  while (pozicevradku and 3)<>0 do zapispixel(0); {pripadne vyplnove byty pro zarovnani delky radku na nasobek 4}
  end;
 zapisbuffer;

 close(soubor);
 writeln(#13'[>>>>>>>>>>]');
 writeln('Hotovo, vysledek najdete v souboru ',jmenosouboru,'.');
 writeln;
 end {konec vypoctu}
else if upcase(nacitadlo[1])='N' then {navod}
 begin
 writeln(' Vyznam vstupnich hodnot je uveden na prilozenem nakresu SCHEMA.GIF.');
 writeln('Hodnoty D, V, fi a psi mohou byt i zaporne. Puntiky na jejich kotach oznacuji');
 writeln('nulovy bod a sipky (tak, jak je vidite na schematu) predstavuji kladny smer.');
 writeln('Ostatni rozmery zaporne byt nemohou (ale nekontroluje se to, tak bacha).');
 writeln(' Pri zadavani hodnot se ve <spicatych zavorkach> zobrazuje vychozi hodnota');
 writeln('ulozena od minula. Bud jenom zmacknete Enter a tim ji znovu pouzijete, nebo');
 writeln('napiste novou. Desetinna mista vzdy oddelujte teckou, ne carkou!');
 writeln(' Vystupem z programu je plast valce 1 ulozeny ve formatu BMP. Rozliseni (DPI)');
 writeln('si zadejte podle moznosti sve tiskarny; cim vic, tim jemnejsi cary, ale take');
 writeln('vetsi objem souboru a delsi doba vypoctu.');
 writeln(' Tolerance okraje diry znamena, jak blizko (v pixelech) musi byt dany bod od');
 writeln('vypoctene hrany pruniku, aby byl povazovan za bod hrany a prislusne vybarven.');
 writeln(' Pri hodnote 0.5 ziskate obvykle caru tlustou 1 px, pri vetsi tlustsi, pri');
 writeln('mensi nesouvislou a pri 0 zadnou (pak nezapomente pro vnitrek diry nastavit');
 writeln('jinou barvu nez pro zbytek plaste, nebo nic neuvidite). Pri urcitych');
 writeln('kombinacich rozmeru a poloh valcu obrysova cara v nekterych mistech neprijemne');
 writeln('tloustne, proto je v takovych pripadech lepsi zadat nulovou toleranci.');
 writeln(' Uhlem psi si zvolite polohu podelneho spoje plaste. Pri malych dirach je');
 writeln('vhodne dat ho nekam mimo ne, pri hodne velkych naopak doprostred nich, aby se');
 writeln('valec dobre slepoval.');
 writeln(' Podel leveho nebo praveho svisleho okraje vytisteneho plaste si prikreslete');
 writeln('spojovaci chlopen, podel horniho a dolniho pak zoubky na prilepeni podstav');
 writeln('(tedy pokud je nelepite natupo). Podstavy si dorysujte rucne.');
 writeln;
 write(' Stisknete Enter pro navrat na hlavni nabidku.');
 readln;
 end
else begin {koncime}
     writeln('Nashledanou.');
     break;
     end;
for b:=1 to 5 do writeln;
Until false;

{$ifdef ukladani}
{ulozeni zadanych hodnot pro priste:}
d1:=2*r1; d2:=2*r2; fi:=fi*180/pi; psi:=psi*180/pi;
konf.uloz(cfgsoubor);
konf.zrus;
{$endif}

writeln;
writeln(' (cc) Mircosoft 2010/2021   http://mircosoft.mzf.cz');
END.

{Princip hledani pruniku dvou valcovych ploch:
-----------------------------------------------
 Vezmu bod na plasti prvniho valce a podivam se, jak je daleko od osy druheho
valce. Kdyz je bliz nez jeho polomer, znamena to, ze je uvnitr nej, tj. uvnitr
oblasti, ktera se ma z plaste prvniho valce vyriznout. Kdyz je dal, je mimo
nej. Kdyz je presne na tom polomeru (plus minus nejaka tolerance, protoze
pracujeme s realnymi cisly), je to bod hranice pruniku.
 Postupuji metodou hrube sily, cili testuji kazdy pixel na plasti prvniho
valce.

 Princip vypoctu vzdalenosti bodu od osy valce:
------------------------------------------------
 Technicky vzato jde o nalezeni vzdalenosti bodu od primky. Pouzil jsem
nasledujici postup (existuji i jine, tenhle se mi zrovna libil):
 Vezmu si rovinu kolmou k ose druheho valce a prochazejici zkoumanym bodem.
Najdu prusecik teto roviny a osy druheho valce. Vzdalenost tohoto pruseciku
od zkoumaneho bodu je to, co hledam.
 Implementace:
1) Zvolim vhodny souradny system. Pocatek jsem dal do stredu te podstavy
   prvniho valce, od ktereho merime vzdalenost d1, osa z miri podel osy
   prvniho valce, osa y ve smeru nejkratsi spojnice obou os a x je jasna.
2) Vypocitam souradnice zkoumaneho bodu (oznacme ho treba A) v prostoru:
                        alfa = o*25.4/(r1*DPI)
                        xA = r1*cos(alfa)
                        yA = r1*sin(alfa)
                        zA = a*25.4/DPI
   Cili: obvodovou souradnici o [px] prevedu na milimetry (*25.4/DPI) a
   vydelim polomerem r1, tim ziskam uhlovou souradnici bodu v rovine x-y.
   Pomoci sin a cos prevedu polarni souradnice (r1,alfa) na (xA,yA).
   zA je opet jenom podelna souradnice a v pixelech prevedena na milimetry.
3) Vyjadrim si osu 2. valce v parametrickem tvaru:
                      x = tg(fi)*t - d*tg(fi)
                      y =               v
                      z =      1*t
   Kde t je realny parametr.
4) Smerovy vektor osy je (tg(fi),0,1) ma byt zaroven normalovy vektor kolme
   roviny, takze jeji obecna rovnice bude:
                    tg(fi)*x + 0*y + z + DD = 0
   DD je konstanta, kterou ziskame dosazenim (xA,yA,zA) za (x,y,z) do teto
   rovnice (chceme, aby rovina tim bodem prochazela).
5) Prusecik roviny a osy ziskame dosazenim parametrickeho vyjadreni osy
   za (x,y,z) v rovnici roviny a z toho vypocitame hodnotu parametru t.
   Tu kdyz dosadime do rovnice osy, ziskame souradnice pruseciku (xP,yP,zP).
6) Spocitame vzdalenost tohoto pruseciku od zkoumaneho bodu:
         vzdalenost = sqrt( sqr(xA-xP) + sqr(yA-yP) + sqr(zA-zP) )
   A je to.
7) Dale je potreba specialne osetrit pripad, kdy jsou na sebe osy valcu kolme,
   protoze by pak tg(fi) utikalo do nekonecna. Rovnice roviny kolme k ose
   druheho valce se zredukuje na:
                                 x = xA
   Zbytek je vicemene stejny, staci jenom prislusne upravit vsechny navazujici
   rovnice.}
